package com.bergerkiller.bukkit.nolagg.meta;

import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftChunk;
import org.bukkit.craftbukkit.CraftWorld;

public class ChunkMetaData {
	public static int unloadDelay;
	public static boolean useUnloadDelay;
	
	public static ChunkMetaData get(Chunk chunk) {
		return get(chunk.getWorld(), chunk.getX(), chunk.getZ());
	}
	public static ChunkMetaData get(final World world, final int x, final int z) {
		return WorldMetaData.get(world).getChunk(x, z);
	}
	public static ChunkMetaData get(final Location location) {
		return WorldMetaData.get(location.getWorld()).getChunk(location);
	}
	
	public static long getKey(int x, int z) {
		return (long) x + 2147483647L | (long) z + 2147483647L << 32;
	}
	
	protected ChunkMetaData(final WorldMetaData world, final int x, int z) {
		this.world = world;
		this.x = x;
		this.z = z;
		this.coord = getKey(x, z);
	}
	private final WorldMetaData world;
	private final int x, z;
	public final long coord;
	
	public long lastAccessTime;
	private short unloadCounter = 0;
	
	public WorldMetaData getMetaWorld() {
		return this.world;
	}
	public CraftWorld getWorld() {
		return this.world.world;
	}
	public WorldServer getNativeWorld() {
		return this.world.getNative();
	}
	public net.minecraft.server.Chunk getNativeChunk() {
		return this.getChunk().getHandle();
	}
	public CraftChunk getChunk() {
		return (CraftChunk) this.getWorld().getChunkAt(this.x, this.z);
	}
	
	public int getX() {
		return this.x;
	}
	public int getZ() {
		return this.z;
	}
	
	public boolean isLoaded() {
		return this.getWorld().isChunkLoaded(this.getX(), this.getZ());
	}
	
	public void touch() {
		this.lastAccessTime = System.currentTimeMillis();
	}
	
	public void markNotUnloadable() {
		this.unloadCounter = 5; //ignore this chunk for 5 update cycles
	}
	
	public boolean isExpired() {
		return this.isExpired(System.currentTimeMillis() - unloadDelay);
	}
	public boolean isExpired(final long time) {
		if (!useUnloadDelay) return true;
		return this.lastAccessTime <= time;
	}
	
	public boolean isPlayerNear() {
		return this.world.isPlayerNear(this);
	}
	
	public boolean isNearSpawnChunk(WorldServer world) {
		return this.isNearChunk(WorldMetaData.getSpawnChunkX(world), WorldMetaData.getSpawnChunkZ(world));
	}
	public boolean isNearChunk(int cx, int cz) {
		return Math.abs(cx - this.x) <= 12 && Math.abs(cz - this.z) <= 12;
	}
	
	public boolean isUnloadable() {
		return this.isUnloadable(true);
	}
	public boolean isUnloadable(boolean checkCounter) {
		WorldServer world = this.getNativeWorld();
		if (world.savingDisabled) return false;
		if (checkCounter && this.unloadCounter > 0) {
			--unloadCounter;
			return false;
		} else {
			return !world.keepSpawnInMemory || !this.isNearSpawnChunk(world);
		}
	}

	public String toString() {
		return "[" + this.getX() + "/" + this.getZ() + "/" + this.getWorld().getName() + "]";
	}
		
	/*
	 * Called when this chunk meta data is being removed (unloaded)
	 */
	public void remove() {
		this.remove(true);
	}
	protected void remove(final boolean removefromworldmeta) {
		//respawn items
		if (removefromworldmeta) this.world.removeChunk(this);
	}
	
	public boolean isChunk(Chunk chunk) {
		if (chunk.getWorld() != this.getWorld()) return false;
		if (chunk.getX() != this.getX()) return false;
		if (chunk.getZ() != this.getZ()) return false;
		return true;
	}

}
